/// <reference types="node" />
import { Browser, ClientSample, Engine, IceLocalCandidate, IceRemoteCandidate, MediaCodecStats, MediaDevice, OperationSystem, Platform } from '@observertc/sample-schemas-js';
import { ObservedCall } from './ObservedCall';
import { EventEmitter } from 'events';
import { ObservedPeerConnection } from './ObservedPeerConnection';
import { PartialBy } from './common/utils';
import { ObservedSfu } from './ObservedSfu';
import { ClientIssue } from './monitors/CallSummary';
import { CallEventReport } from '@observertc/report-schemas-js';
import { CalculatedScore } from './common/CalculatedScore';
export type ObservedClientModel = {
    clientId: string;
    mediaUnitId: string;
    userId?: string;
    marker?: string;
    operationSystem?: OperationSystem;
    engine?: Engine;
    platform?: Platform;
    browser?: Browser;
    coordinates?: {
        latitude: number;
        longitude: number;
    };
};
export type ObservedClientEvents = {
    update: [
        {
            sample: ClientSample;
            elapsedTimeInMs: number;
        }
    ];
    close: [];
    joined: [];
    left: [];
    newpeerconnection: [ObservedPeerConnection];
    iceconnectionstatechange: [
        {
            peerConnection: ObservedPeerConnection;
            state: string;
        }
    ];
    icegatheringstatechange: [
        {
            peerConnection: ObservedPeerConnection;
            state: string;
        }
    ];
    connectionstatechange: [
        {
            peerConnection: ObservedPeerConnection;
            state: string;
        }
    ];
    selectedcandidatepair: [
        {
            peerConnection: ObservedPeerConnection;
            localCandidate: IceLocalCandidate;
            remoteCandidate: IceRemoteCandidate;
        }
    ];
    issue: [ClientIssue];
    usingturn: [boolean];
    usermediaerror: [string];
    rejoined: [
        {
            lastJoined: number;
        }
    ];
    score: [CalculatedScore];
};
export declare interface ObservedClient<AppData extends Record<string, unknown> = Record<string, unknown>> {
    on<U extends keyof ObservedClientEvents>(event: U, listener: (...args: ObservedClientEvents[U]) => void): this;
    off<U extends keyof ObservedClientEvents>(event: U, listener: (...args: ObservedClientEvents[U]) => void): this;
    once<U extends keyof ObservedClientEvents>(event: U, listener: (...args: ObservedClientEvents[U]) => void): this;
    emit<U extends keyof ObservedClientEvents>(event: U, ...args: ObservedClientEvents[U]): boolean;
    readonly appData: AppData;
}
type PendingPeerConnectionTimestamp = {
    type: 'opened' | 'closed';
    peerConnectionId: string;
    timestamp: number;
};
type PendingMediaTrackTimestamp = {
    type: 'added' | 'removed';
    peerConnectionId: string;
    mediaTrackId: string;
    timestamp: number;
};
export declare class ObservedClient<AppData extends Record<string, unknown> = Record<string, unknown>> extends EventEmitter {
    private readonly _model;
    readonly call: ObservedCall;
    readonly appData: AppData;
    readonly created: number;
    updated: number;
    sfuId?: string;
    private readonly _peerConnections;
    private _closed;
    private _acceptedSamples;
    private _timeZoneOffsetInHours?;
    private _joined?;
    private _left?;
    lastStatsTimestamp?: number;
    score?: CalculatedScore;
    usingTURN: boolean;
    availableOutgoingBitrate: number;
    availableIncomingBitrate: number;
    totalInboundPacketsLost: number;
    totalInboundPacketsReceived: number;
    totalOutboundPacketsSent: number;
    totalDataChannelBytesSent: number;
    totalDataChannelBytesReceived: number;
    totalSentBytes: number;
    totalReceivedBytes: number;
    totalReceivedAudioBytes: number;
    totalReceivedVideoBytes: number;
    totalSentAudioBytes: number;
    totalSentVideoBytes: number;
    deltaReceivedAudioBytes: number;
    deltaReceivedVideoBytes: number;
    deltaSentAudioBytes: number;
    deltaSentVideoBytes: number;
    deltaDataChannelBytesSent: number;
    deltaDataChannelBytesReceived: number;
    deltaInboundPacketsLost: number;
    deltaInboundPacketsReceived: number;
    deltaOutboundPacketsSent: number;
    avgRttInMs?: number;
    outboundAudioBitrate?: number;
    outboundVideoBitrate?: number;
    inboundAudioBitrate?: number;
    inboundVideoBitrate?: number;
    mediaConstraints: string[];
    readonly mediaDevices: MediaDevice[];
    readonly mediaCodecs: MediaCodecStats[];
    readonly userMediaErrors: string[];
    issues: ClientIssue[];
    ωpendingPeerConnectionTimestamps: PendingPeerConnectionTimestamp[];
    ωpendingMediaTrackTimestamps: PendingMediaTrackTimestamp[];
    constructor(_model: ObservedClientModel, call: ObservedCall, appData: AppData);
    getSfu<T extends Record<string, unknown> = Record<string, unknown>>(): ObservedSfu<T> | undefined;
    get joined(): number | undefined;
    set joined(value: number | undefined);
    get left(): number | undefined;
    set left(value: number | undefined);
    get sfu(): ObservedSfu | undefined;
    get coordinates(): {
        latitude: number;
        longitude: number;
    } | undefined;
    set coordinates(value: {
        latitude: number;
        longitude: number;
    } | undefined);
    get clientId(): string;
    get roomId(): string;
    get serviceId(): string;
    get callId(): string;
    get mediaUnitId(): string;
    get reports(): import("./ReportsCollector").ReportsCollector;
    get userId(): string | undefined;
    set userId(userId: string | undefined);
    get timeZoneOffsetInHours(): number | undefined;
    get marker(): string | undefined;
    get browser(): Browser | undefined;
    get engine(): Engine | undefined;
    get operationSystem(): OperationSystem | undefined;
    get platform(): Platform | undefined;
    get acceptedSamples(): number;
    get closed(): boolean;
    get peerConnections(): ReadonlyMap<string, ObservedPeerConnection>;
    close(): void;
    addEventReport(params: PartialBy<Omit<CallEventReport, 'serviceId' | 'roomId' | 'callId' | 'clientId' | 'userId' | 'marker' | 'attachments'>, 'timestamp'> & {
        attachments?: Record<string, unknown>;
    }): void;
    addExtensionStatsReport(extensionType: string, payload?: Record<string, unknown>): void;
    addIssue(issue: ClientIssue): void;
    accept(sample: ClientSample): void;
    private _addAndEmitIssue;
    private _createPeerConnection;
}
export {};
//# sourceMappingURL=ObservedClient.d.ts.map