"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObservedICE = void 0;
const events_1 = require("events");
const CallMetaReports_1 = require("./common/CallMetaReports");
class ObservedICE extends events_1.EventEmitter {
    static create(peerConnection) {
        return new ObservedICE(peerConnection);
    }
    constructor(peerConnection) {
        super();
        this.peerConnection = peerConnection;
        this._localCandidates = new Map();
        this._remoteCandidates = new Map();
        this.usingTURN = false;
        this.deltaBytesReceived = 0;
        this.deltaBytesSent = 0;
        this.deltaPacketsReceived = 0;
        this.deltaPacketsSent = 0;
        this.totalBytesReceived = 0;
        this.totalBytesSent = 0;
        this.totalPacketsReceived = 0;
        this.totalPacketsSent = 0;
        this._updated = Date.now();
        this._closed = false;
        this.setMaxListeners(Infinity);
    }
    get reports() {
        return this.peerConnection.reports;
    }
    get localCandidates() {
        return this._localCandidates;
    }
    get remoteCandidates() {
        return this._remoteCandidates;
    }
    get selectedLocalCandidate() {
        var _a;
        return this._localCandidates.get((_a = this._selectedLocalCandidateId) !== null && _a !== void 0 ? _a : '');
    }
    get selectedRemoteCandidate() {
        var _a;
        return this._remoteCandidates.get((_a = this._selectedRemoteCandidateId) !== null && _a !== void 0 ? _a : '');
    }
    get stats() {
        return this._stats;
    }
    get updated() {
        return this._updated;
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this._closed)
            return;
        this._closed = true;
        this.emit('close');
    }
    addLocalCandidate(candidate, sampleTimestamp) {
        if (!candidate.id)
            return;
        const newCandidate = !this._localCandidates.has(candidate.id);
        this._localCandidates.set(candidate.id, candidate);
        if (newCandidate) {
            const callMetaReport = (0, CallMetaReports_1.createCallMetaReport)(this.peerConnection.client.serviceId, this.peerConnection.client.mediaUnitId, this.peerConnection.client.roomId, this.peerConnection.client.callId, this.peerConnection.client.clientId, {
                type: CallMetaReports_1.CallMetaType.ICE_LOCAL_CANDIDATE,
                payload: candidate,
            }, this.peerConnection.client.userId, this.peerConnection.peerConnectionId, sampleTimestamp);
            this.reports.addCallMetaReport(callMetaReport);
            this.emit('new-local-candidate', candidate);
        }
    }
    addRemoteCandidate(candidate, sampleTimestamp) {
        if (!candidate.id)
            return;
        const newCandidate = !this._remoteCandidates.has(candidate.id);
        this._remoteCandidates.set(candidate.id, candidate);
        if (newCandidate) {
            const callMetaReport = (0, CallMetaReports_1.createCallMetaReport)(this.peerConnection.client.serviceId, this.peerConnection.client.mediaUnitId, this.peerConnection.client.roomId, this.peerConnection.client.callId, this.peerConnection.client.clientId, {
                type: CallMetaReports_1.CallMetaType.ICE_REMOTE_CANDIDATE,
                payload: candidate,
            }, this.peerConnection.client.userId, this.peerConnection.peerConnectionId, sampleTimestamp);
            this.reports.addCallMetaReport(callMetaReport);
            this.emit('new-remote-candidate', candidate);
        }
    }
    resetMetrics() {
        this.deltaBytesReceived = 0;
        this.deltaBytesSent = 0;
        this.deltaPacketsReceived = 0;
        this.deltaPacketsSent = 0;
        this.currentRttInMs = undefined;
    }
    update(candidatePair, timestamp) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r;
        const now = Date.now();
        const elapsedTimeInMs = now - this._updated;
        const report = Object.assign(Object.assign({ serviceId: this.peerConnection.client.serviceId, mediaUnitId: this.peerConnection.client.mediaUnitId, roomId: this.peerConnection.client.roomId, callId: this.peerConnection.client.callId, clientId: this.peerConnection.client.clientId, userId: this.peerConnection.client.userId, timestamp }, candidatePair), { sampleSeq: -1, marker: this.marker });
        this.reports.addIceCandidatePairReport(report);
        if (!candidatePair.nominated)
            return;
        this.deltaBytesReceived = ((_a = candidatePair.bytesReceived) !== null && _a !== void 0 ? _a : 0) - ((_c = (_b = this._stats) === null || _b === void 0 ? void 0 : _b.bytesReceived) !== null && _c !== void 0 ? _c : 0);
        this.deltaBytesSent = ((_d = candidatePair.bytesSent) !== null && _d !== void 0 ? _d : 0) - ((_f = (_e = this._stats) === null || _e === void 0 ? void 0 : _e.bytesSent) !== null && _f !== void 0 ? _f : 0);
        this.deltaPacketsReceived = ((_g = candidatePair.packetsReceived) !== null && _g !== void 0 ? _g : 0) - ((_j = (_h = this._stats) === null || _h === void 0 ? void 0 : _h.packetsReceived) !== null && _j !== void 0 ? _j : 0);
        this.deltaPacketsSent = ((_k = candidatePair.packetsSent) !== null && _k !== void 0 ? _k : 0) - ((_m = (_l = this._stats) === null || _l === void 0 ? void 0 : _l.packetsSent) !== null && _m !== void 0 ? _m : 0);
        this.totalBytesReceived += this.deltaBytesReceived;
        this.totalBytesSent += this.deltaBytesSent;
        this.totalPacketsReceived += this.deltaPacketsReceived;
        this.totalPacketsSent += this.deltaPacketsSent;
        this.currentRttInMs = candidatePair.currentRoundTripTime ? candidatePair.currentRoundTripTime * 1000 : undefined;
        this._stats = candidatePair;
        if (candidatePair.localCandidateId &&
            candidatePair.localCandidateId !== this._selectedLocalCandidateId &&
            this._localCandidates.has(candidatePair.localCandidateId) &&
            candidatePair.remoteCandidateId &&
            candidatePair.remoteCandidateId !== this._selectedRemoteCandidateId &&
            this._remoteCandidates.has(candidatePair.remoteCandidateId)) {
            this._selectedLocalCandidateId = candidatePair.localCandidateId;
            this._selectedRemoteCandidateId = candidatePair.remoteCandidateId;
            const localCandidate = this._localCandidates.get(candidatePair.localCandidateId);
            const remoteCandidate = this._remoteCandidates.get(candidatePair.remoteCandidateId);
            if (localCandidate && remoteCandidate) {
                const wasUsingTURN = this.usingTURN;
                this.usingTURN = ((_p = ((_o = localCandidate.candidateType) === null || _o === void 0 ? void 0 : _o.toLocaleLowerCase()) === 'relay') !== null && _p !== void 0 ? _p : false) && ((_r = (_q = localCandidate.url) === null || _q === void 0 ? void 0 : _q.startsWith('turn:')) !== null && _r !== void 0 ? _r : false);
                this.emit('new-selected-candidate-pair', {
                    localCandidate,
                    remoteCandidate,
                });
                if (wasUsingTURN !== this.usingTURN) {
                    this.emit('usingturnchanged', this.usingTURN);
                }
            }
        }
        this._updated = now;
        this.emit('update', {
            elapsedTimeInMs,
        });
    }
}
exports.ObservedICE = ObservedICE;
