"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObservedInboundVideoTrack = void 0;
const events_1 = require("events");
const CalculatedScore_1 = require("./common/CalculatedScore");
class ObservedInboundVideoTrack extends events_1.EventEmitter {
    constructor(_model, peerConnection) {
        super();
        this._model = _model;
        this.peerConnection = peerConnection;
        this.created = Date.now();
        this.visited = false;
        this.contentType = 'standard';
        this._stats = new Map();
        this._closed = false;
        this._updated = Date.now();
        this._lastMaxStatsTimestamp = 0;
        this.bitrate = 0;
        this.fractionLoss = 0;
        this.totalLostPackets = 0;
        this.totalReceivedPackets = 0;
        this.totalBytesReceived = 0;
        this.totalReceivedFrames = 0;
        this.totalDecodedFrames = 0;
        this.totalDroppedFrames = 0;
        this.deltaLostPackets = 0;
        this.deltaReceivedPackets = 0;
        this.deltaBytesReceived = 0;
        this.deltaReceivedFrames = 0;
        this.deltaDecodedFrames = 0;
        this.deltaDroppedFrames = 0;
        this.ωpendingIssuesForScores = [];
        this.setMaxListeners(Infinity);
    }
    get serviceId() {
        return this.peerConnection.serviceId;
    }
    get roomId() {
        return this.peerConnection.roomId;
    }
    get callId() {
        return this.peerConnection.callId;
    }
    get clientId() {
        return this.peerConnection.clientId;
    }
    get mediaUnitId() {
        return this.peerConnection.mediaUnitId;
    }
    get peerConnectionId() {
        return this.peerConnection.peerConnectionId;
    }
    get trackId() {
        return this._model.trackId;
    }
    get sfuStreamId() {
        return this._model.sfuStreamId;
    }
    get sfuSinkId() {
        return this._model.sfuSinkId;
    }
    get updated() {
        return this._updated;
    }
    get statsTimestamp() {
        return this._lastMaxStatsTimestamp;
    }
    get stats() {
        return this._stats;
    }
    get reports() {
        return this.peerConnection.reports;
    }
    set remoteOutboundTrack(track) {
        if (this._closed)
            return;
        if (this._remoteOutboundTrack)
            return;
        if (!track)
            return;
        track.once('close', () => {
            this._remoteOutboundTrack = undefined;
            this.emit('remoteoutboundtrack', undefined);
        });
        this._remoteOutboundTrack = track;
        this.emit('remoteoutboundtrack', track);
    }
    get remoteOutboundTrack() {
        return this._remoteOutboundTrack;
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this._closed)
            return;
        this._closed = true;
        this.emit('close');
    }
    update(sample, statsTimestamp) {
        var _a, _b, _c, _d, _e;
        if (this._closed)
            return;
        const now = Date.now();
        const report = Object.assign(Object.assign({ serviceId: this.peerConnection.client.call.serviceId, roomId: this.peerConnection.client.call.roomId, callId: this.peerConnection.client.call.callId, clientId: this.peerConnection.client.clientId, userId: this.peerConnection.client.userId, mediaUnitId: this.peerConnection.client.mediaUnitId, peerConnectionId: this.peerConnection.peerConnectionId }, sample), { timestamp: statsTimestamp, sampleSeq: -1, remoteClientId: (_a = sample.remoteClientId) !== null && _a !== void 0 ? _a : (_b = this.remoteOutboundTrack) === null || _b === void 0 ? void 0 : _b.peerConnection.client.clientId, remoteUserId: (_c = this.remoteOutboundTrack) === null || _c === void 0 ? void 0 : _c.peerConnection.client.userId, remoteTrackId: (_d = this.remoteOutboundTrack) === null || _d === void 0 ? void 0 : _d.trackId, remotePeerConnectionId: (_e = this.remoteOutboundTrack) === null || _e === void 0 ? void 0 : _e.peerConnection.peerConnectionId, marker: this.marker });
        this.reports.addInboundVideoTrackReport(report);
        if (!this._model.sfuStreamId && sample.sfuStreamId) {
            this._model.sfuStreamId = sample.sfuStreamId;
        }
        if (!this._model.sfuSinkId && sample.sfuSinkId) {
            this._model.sfuSinkId = sample.sfuSinkId;
        }
        const elapsedTimeInMs = Math.max(1, now - this._updated);
        const lastStat = this._stats.get(sample.ssrc);
        const rttInMs = sample.roundTripTime ? sample.roundTripTime * 1000 : undefined;
        let bitrate = 0;
        let deltaReceivedBytes = 0;
        let fractionLost = 0;
        let deltaLostPackets = 0;
        let deltaReceivedPackets = 0;
        if ((lastStat === null || lastStat === void 0 ? void 0 : lastStat.bytesReceived) && sample.bytesReceived && lastStat.bytesReceived < sample.bytesReceived) {
            deltaReceivedBytes = sample.bytesReceived - lastStat.bytesReceived;
            bitrate = deltaReceivedBytes / (elapsedTimeInMs / 1000);
        }
        if ((lastStat === null || lastStat === void 0 ? void 0 : lastStat.packetsReceived) && (sample === null || sample === void 0 ? void 0 : sample.packetsReceived) && lastStat.packetsReceived < sample.packetsReceived) {
            deltaReceivedPackets = sample.packetsReceived - lastStat.packetsReceived;
        }
        if ((lastStat === null || lastStat === void 0 ? void 0 : lastStat.packetsLost) && sample.packetsLost && lastStat.packetsLost < sample.packetsLost) {
            deltaLostPackets = sample.packetsLost - lastStat.packetsLost;
            if (0 < deltaReceivedPackets) {
                fractionLost = deltaLostPackets / (deltaReceivedPackets + deltaLostPackets);
            }
        }
        let deltaDecodedFrames;
        let deltaDroppedFrames;
        let deltaReceivedFrames;
        const videoSample = sample;
        const lastVideoStat = lastStat;
        if (videoSample.framesDecoded && (lastVideoStat === null || lastVideoStat === void 0 ? void 0 : lastVideoStat.framesDecoded) && lastVideoStat.framesDecoded < videoSample.framesDecoded) {
            deltaDecodedFrames = videoSample.framesDecoded - lastVideoStat.framesDecoded;
        }
        if (videoSample.framesDropped && (lastVideoStat === null || lastVideoStat === void 0 ? void 0 : lastVideoStat.framesDropped) && lastVideoStat.framesDropped < videoSample.framesDropped) {
            deltaDroppedFrames = videoSample.framesDropped - lastVideoStat.framesDropped;
        }
        if (videoSample.framesReceived && (lastVideoStat === null || lastVideoStat === void 0 ? void 0 : lastVideoStat.framesReceived) && lastVideoStat.framesReceived < videoSample.framesReceived) {
            deltaReceivedFrames = videoSample.framesReceived - lastVideoStat.framesReceived;
        }
        const stats = Object.assign(Object.assign({}, sample), { fractionLost,
            rttInMs,
            bitrate, ssrc: sample.ssrc, deltaReceivedBytes,
            deltaLostPackets,
            deltaReceivedPackets,
            deltaReceivedFrames,
            deltaDecodedFrames,
            deltaDroppedFrames,
            statsTimestamp });
        this._stats.set(sample.ssrc, stats);
        this.visited = true;
        // a peer connection is active if it has at least one active track
        this.peerConnection.visited = true;
        this._updated = now;
        this.emit('update', {
            elapsedTimeInMs,
        });
    }
    updateMetrics() {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        let maxStatsTimestamp = 0;
        let rttInMsSum = 0;
        let jitterSum = 0;
        let size = 0;
        this.bitrate = 0;
        this.rttInMs = undefined;
        this.deltaBytesReceived = 0;
        this.deltaLostPackets = 0;
        this.deltaReceivedFrames = 0;
        this.deltaDecodedFrames = 0;
        this.deltaDroppedFrames = 0;
        this.fractionLoss = 0;
        const highestLayer = { ssrc: -1, bitrate: -1 };
        for (const [ssrc, stats] of this._stats) {
            if (stats.statsTimestamp <= this._lastMaxStatsTimestamp)
                continue;
            this.deltaBytesReceived += (_a = stats.deltaReceivedBytes) !== null && _a !== void 0 ? _a : 0;
            this.deltaLostPackets += (_b = stats.deltaLostPackets) !== null && _b !== void 0 ? _b : 0;
            this.deltaReceivedPackets += (_c = stats.deltaReceivedFrames) !== null && _c !== void 0 ? _c : 0;
            this.deltaReceivedFrames += (_d = stats.deltaReceivedFrames) !== null && _d !== void 0 ? _d : 0;
            this.deltaDecodedFrames += (_e = stats.deltaDecodedFrames) !== null && _e !== void 0 ? _e : 0;
            this.deltaDroppedFrames += (_f = stats.deltaDroppedFrames) !== null && _f !== void 0 ? _f : 0;
            this.bitrate += stats.bitrate;
            maxStatsTimestamp = Math.max(maxStatsTimestamp, stats.statsTimestamp);
            if (highestLayer.bitrate < stats.bitrate) {
                highestLayer.ssrc = ssrc;
            }
            rttInMsSum += (_g = stats.rttInMs) !== null && _g !== void 0 ? _g : 0;
            jitterSum += (_h = stats.jitter) !== null && _h !== void 0 ? _h : 0;
            ++size;
        }
        this.highestLayer = this.stats.get(highestLayer.ssrc);
        this.totalBytesReceived += this.deltaBytesReceived;
        this.totalLostPackets += this.deltaLostPackets;
        this.totalReceivedPackets += this.deltaReceivedPackets;
        this.totalReceivedFrames += this.deltaReceivedFrames;
        this.totalDecodedFrames += this.deltaDecodedFrames;
        this.totalDroppedFrames += this.deltaDroppedFrames;
        this.rttInMs = rttInMsSum / Math.max(size, 1);
        this.jitter = jitterSum / Math.max(size, 1);
        this.fractionLoss = 0 < this.deltaReceivedPackets && 0 < this.deltaLostPackets ? (this.deltaLostPackets / (this.deltaReceivedPackets + this.deltaLostPackets)) : 0;
        this._lastMaxStatsTimestamp = maxStatsTimestamp;
        this._updateQualityScore(maxStatsTimestamp);
    }
    _updateQualityScore(timestamp) {
        const newIssues = this.ωpendingIssuesForScores;
        const score = (0, CalculatedScore_1.calculateBaseVideoScore)(this, newIssues);
        if (0 < this.ωpendingIssuesForScores.length) {
            this.ωpendingIssuesForScores = [];
        }
        if (!score)
            return (this.score = undefined);
        if (0.05 < this.fractionLoss) {
            score.score *= 0.5;
            score.remarks.push({
                severity: 'major',
                text: 'Fraction loss is too high',
            });
        }
        score.timestamp = timestamp;
        this.score = score;
        this.emit('score', this.score);
    }
}
exports.ObservedInboundVideoTrack = ObservedInboundVideoTrack;
