"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObservedOutboundAudioTrack = void 0;
const events_1 = require("events");
const CalculatedScore_1 = require("./common/CalculatedScore");
class ObservedOutboundAudioTrack extends events_1.EventEmitter {
    constructor(_model, peerConnection) {
        super();
        this._model = _model;
        this.peerConnection = peerConnection;
        this.created = Date.now();
        this.visited = false;
        this.bitrate = 0;
        this.totalLostPackets = 0;
        this.totalSentPackets = 0;
        this.totalSentBytes = 0;
        this.totalSentFrames = 0;
        this.deltaLostPackets = 0;
        this.deltaSentPackets = 0;
        this.deltaSentBytes = 0;
        this.deltaSentFrames = 0;
        this.deltaEncodedFrames = 0;
        this.sendingBitrate = 0;
        this._stats = new Map();
        this._lastMaxStatsTimestamp = 0;
        this._closed = false;
        this._updated = Date.now();
        this.ωpendingIssuesForScores = [];
        this.remoteInboundTracks = new Map();
        this.setMaxListeners(Infinity);
    }
    get serviceId() {
        return this.peerConnection.serviceId;
    }
    get roomId() {
        return this.peerConnection.roomId;
    }
    get callId() {
        return this.peerConnection.callId;
    }
    get clientId() {
        return this.peerConnection.clientId;
    }
    get mediaUnitId() {
        return this.peerConnection.mediaUnitId;
    }
    get peerConnectionId() {
        return this.peerConnection.peerConnectionId;
    }
    get trackId() {
        return this._model.trackId;
    }
    get sfuStreamId() {
        return this._model.sfuStreamId;
    }
    get reports() {
        return this.peerConnection.reports;
    }
    get statsTimestamp() {
        return this._lastMaxStatsTimestamp;
    }
    get updated() {
        return this._updated;
    }
    get stats() {
        return this._stats;
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this._closed)
            return;
        this._closed = true;
        this.emit('close');
    }
    update(sample, statsTimestamp) {
        if (this._closed)
            return;
        const now = Date.now();
        const report = Object.assign(Object.assign({ serviceId: this.peerConnection.client.serviceId, roomId: this.peerConnection.client.call.roomId, callId: this.peerConnection.client.call.callId, clientId: this.peerConnection.client.clientId, userId: this.peerConnection.client.userId, mediaUnitId: this.peerConnection.client.mediaUnitId, peerConnectionId: this.peerConnection.peerConnectionId }, sample), { timestamp: statsTimestamp, sampleSeq: -1, marker: this.marker });
        this.reports.addOutboundAudioTrackReport(report);
        const elapsedTimeInMs = Math.max(1, now - this._updated);
        const lastStat = this._stats.get(sample.ssrc);
        const rttInMs = sample.roundTripTime ? sample.roundTripTime * 1000 : undefined;
        let bitrate = 0;
        let deltaLostPackets = 0;
        let deltaSentPackets = 0;
        let deltaSentBytes = 0;
        if (sample.bytesSent && (lastStat === null || lastStat === void 0 ? void 0 : lastStat.bytesSent) && lastStat.bytesSent < sample.bytesSent) {
            bitrate = (sample.bytesSent - lastStat.bytesSent) * 8 / (elapsedTimeInMs / 1000);
        }
        if (sample.packetsLost && (lastStat === null || lastStat === void 0 ? void 0 : lastStat.packetsLost) && lastStat.packetsLost < sample.packetsLost) {
            deltaLostPackets = sample.packetsLost - lastStat.packetsLost;
        }
        if (sample.packetsSent && (lastStat === null || lastStat === void 0 ? void 0 : lastStat.packetsSent) && lastStat.packetsSent < sample.packetsSent) {
            deltaSentPackets = sample.packetsSent - lastStat.packetsSent;
        }
        if (sample.bytesSent && (lastStat === null || lastStat === void 0 ? void 0 : lastStat.bytesSent) && lastStat.bytesSent < sample.bytesSent) {
            deltaSentBytes = sample.bytesSent - lastStat.bytesSent;
        }
        let deltaEncodedFrames;
        let deltaSentFrames;
        const videoSample = sample;
        const lastVideoStats = lastStat;
        if ((videoSample === null || videoSample === void 0 ? void 0 : videoSample.framesEncoded) && (lastVideoStats === null || lastVideoStats === void 0 ? void 0 : lastVideoStats.framesEncoded) && lastVideoStats.framesEncoded < videoSample.framesEncoded) {
            deltaEncodedFrames = videoSample.framesEncoded - lastVideoStats.framesEncoded;
        }
        if ((videoSample === null || videoSample === void 0 ? void 0 : videoSample.framesSent) && (lastVideoStats === null || lastVideoStats === void 0 ? void 0 : lastVideoStats.framesSent) && lastVideoStats.framesSent < videoSample.framesSent) {
            deltaSentFrames = videoSample.framesSent - lastVideoStats.framesSent;
        }
        if (videoSample.qualityLimitationReason && (lastVideoStats === null || lastVideoStats === void 0 ? void 0 : lastVideoStats.qualityLimitationReason) !== videoSample.qualityLimitationReason) {
            this.emit('qualitylimitationchanged', videoSample.qualityLimitationReason);
        }
        const stats = Object.assign(Object.assign({}, sample), { rttInMs,
            bitrate, ssrc: sample.ssrc, deltaLostPackets,
            deltaSentPackets,
            deltaSentBytes,
            deltaEncodedFrames,
            deltaSentFrames,
            statsTimestamp });
        this._stats.set(sample.ssrc, stats);
        this.visited = true;
        // a peer connection is active if it has at least one active track
        this.peerConnection.visited = true;
        this._updated = now;
        this.emit('update', {
            elapsedTimeInMs,
        });
    }
    updateMetrics() {
        var _a, _b, _c, _d, _e, _f, _g;
        let maxStatsTimestamp = 0;
        let rttInMsSum = 0;
        let jitterSum = 0;
        let size = 0;
        this.bitrate = 0;
        this.rttInMs = undefined;
        this.sendingBitrate = 0;
        this.deltaLostPackets = 0;
        this.deltaSentPackets = 0;
        this.deltaSentBytes = 0;
        this.deltaSentFrames = 0;
        this.deltaEncodedFrames = 0;
        for (const [, stats] of this._stats) {
            if (stats.statsTimestamp <= this._lastMaxStatsTimestamp)
                continue;
            this.deltaLostPackets += (_a = stats.deltaLostPackets) !== null && _a !== void 0 ? _a : 0;
            this.deltaSentPackets += (_b = stats.deltaSentPackets) !== null && _b !== void 0 ? _b : 0;
            this.deltaSentBytes += (_c = stats.deltaSentBytes) !== null && _c !== void 0 ? _c : 0;
            this.deltaSentFrames += (_d = stats.deltaSentFrames) !== null && _d !== void 0 ? _d : 0;
            this.deltaEncodedFrames += (_e = stats.deltaEncodedFrames) !== null && _e !== void 0 ? _e : 0;
            this.bitrate += stats.bitrate;
            maxStatsTimestamp = Math.max(maxStatsTimestamp, stats.statsTimestamp);
            rttInMsSum += (_f = stats.rttInMs) !== null && _f !== void 0 ? _f : 0;
            jitterSum += (_g = stats.jitter) !== null && _g !== void 0 ? _g : 0;
            ++size;
        }
        const now = Date.now();
        if (this._lastUpdateMetrics) {
            this.sendingBitrate = (this.deltaSentBytes * 8) / ((now - this._lastUpdateMetrics) / 1000);
        }
        this._lastUpdateMetrics = now;
        this.totalLostPackets += this.deltaLostPackets;
        this.totalSentPackets += this.deltaSentPackets;
        this.totalSentBytes += this.deltaSentBytes;
        this.totalSentFrames += this.deltaSentFrames;
        this.rttInMs = rttInMsSum / Math.max(size, 1);
        this.jitter = jitterSum / Math.max(size, 1);
        this._lastMaxStatsTimestamp = maxStatsTimestamp;
        this._updateQualityScore(maxStatsTimestamp);
    }
    _updateQualityScore(timestamp) {
        const newIssues = this.ωpendingIssuesForScores;
        const score = (0, CalculatedScore_1.calculateBaseAudioScore)(this, newIssues);
        if (0 < newIssues.length) {
            this.ωpendingIssuesForScores = [];
        }
        if (!score)
            return (this.score = undefined);
        score.timestamp = timestamp;
        this.score = score;
        this.emit('score', this.score);
    }
}
exports.ObservedOutboundAudioTrack = ObservedOutboundAudioTrack;
