/// <reference types="node" />
import { EventEmitter } from 'events';
import { PeerConnectionTransport } from '@observertc/sample-schemas-js';
import { ObservedClient } from './ObservedClient';
import { CallEventReport } from '@observertc/report-schemas-js';
import { ObservedICE } from './ObservedICE';
import { ObservedDataChannel } from './ObservedDataChannel';
import { PartialBy } from './common/utils';
import { ObservedInboundAudioTrack, ObservedInboundAudioTrackModel } from './ObservedInboundAudioTrack';
import { ObservedInboundVideoTrack, ObservedInboundVideoTrackModel } from './ObservedInboundVideoTrack';
import { ObservedOutboundAudioTrack, ObservedOutboundAudioTrackModel } from './ObservedOutboundAudioTrack';
import { ObservedOutboundVideoTrack, ObservedOutboundVideoTrackModel } from './ObservedOutboundVideoTrack';
import { CalculatedScore } from './common/CalculatedScore';
import { ClientIssue } from './monitors/CallSummary';
export type ObservedPeerConnectionEvents = {
    update: [
        {
            elapsedTimeInMs: number;
        }
    ];
    close: [];
    score: [CalculatedScore];
    newinboudaudiotrack: [ObservedInboundAudioTrack];
    newinboudvideotrack: [ObservedInboundVideoTrack];
    newoutboundaudiotrack: [ObservedOutboundAudioTrack];
    newoutboundvideotrack: [ObservedOutboundVideoTrack];
    newdatachannel: [ObservedDataChannel];
};
export type ObservedPeerConnectionModel = {
    peerConnectionId: string;
    label?: string;
};
export type ObservedPeerConnectionStats = Omit<PeerConnectionTransport, 'transportId' | 'label'>;
export declare interface ObservedPeerConnection {
    on<U extends keyof ObservedPeerConnectionEvents>(event: U, listener: (...args: ObservedPeerConnectionEvents[U]) => void): this;
    off<U extends keyof ObservedPeerConnectionEvents>(event: U, listener: (...args: ObservedPeerConnectionEvents[U]) => void): this;
    once<U extends keyof ObservedPeerConnectionEvents>(event: U, listener: (...args: ObservedPeerConnectionEvents[U]) => void): this;
    emit<U extends keyof ObservedPeerConnectionEvents>(event: U, ...args: ObservedPeerConnectionEvents[U]): boolean;
}
export declare class ObservedPeerConnection extends EventEmitter {
    private readonly _model;
    readonly client: ObservedClient;
    readonly created: number;
    visited: boolean;
    opened?: number;
    closedTimestamp?: number;
    private _elapsedTimeSinceLastUpdate?;
    private _statsTimestamp?;
    private _stabilityScores;
    ωpendingIssuesForScores: ClientIssue[];
    score?: CalculatedScore;
    totalInboundPacketsLost: number;
    totalInboundPacketsReceived: number;
    totalOutboundPacketsSent: number;
    totalDataChannelBytesSent: number;
    totalDataChannelBytesReceived: number;
    totalSentAudioBytes: number;
    totalSentVideoBytes: number;
    totalSentAudioPackets: number;
    totalSentVideoPackets: number;
    totalReceivedAudioPacktes: number;
    totalReceivedVideoPackets: number;
    totalReceivedAudioBytes: number;
    totalReceivedVideoBytes: number;
    deltaInboundPacketsLost: number;
    deltaInboundPacketsReceived: number;
    deltaOutboundPacketsSent: number;
    deltaDataChannelBytesSent: number;
    deltaDataChannelBytesReceived: number;
    deltaInboundReceivedBytes: number;
    deltaOutboundSentBytes: number;
    deltaReceivedAudioBytes: number;
    deltaReceivedVideoBytes: number;
    deltaReceivedAudioPackets: number;
    deltaReceivedVideoPackets: number;
    deltaSentAudioBytes: number;
    deltaSentVideoBytes: number;
    receivingPacketsPerSecond: number;
    sendingPacketsPerSecond: number;
    sendingAudioBitrate: number;
    sendingVideoBitrate: number;
    receivingAudioBitrate: number;
    receivingVideoBitrate: number;
    avgRttInMs?: number;
    avgJitter?: number;
    private _closed;
    private _updated;
    private _sample?;
    private _marker?;
    readonly ICE: ObservedICE;
    private readonly _inboundAudioTracks;
    private readonly _inboundVideoTracks;
    private readonly _outboundAudioTracks;
    private readonly _outboundVideoTracks;
    private readonly _dataChannels;
    constructor(_model: ObservedPeerConnectionModel, client: ObservedClient);
    get serviceId(): string;
    get roomId(): string;
    get callId(): string;
    get clientId(): string;
    get mediaUnitId(): string;
    get label(): string | undefined;
    set label(value: string | undefined);
    get usingTURN(): boolean;
    get availableOutgoingBitrate(): number | undefined;
    get availableIncomingBitrate(): number | undefined;
    get peerConnectionId(): string;
    get reports(): import("./ReportsCollector").ReportsCollector;
    get stats(): ObservedPeerConnectionStats | undefined;
    get updated(): number;
    get inboundAudioTracks(): ReadonlyMap<string, ObservedInboundAudioTrack>;
    get inboundVideoTracks(): ReadonlyMap<string, ObservedInboundVideoTrack>;
    get outboundAudioTracks(): ReadonlyMap<string, ObservedOutboundAudioTrack>;
    get outboundVideoTracks(): ReadonlyMap<string, ObservedOutboundVideoTrack>;
    get dataChannels(): ReadonlyMap<number, ObservedDataChannel>;
    get uptimeInMs(): number;
    get marker(): string | undefined;
    set marker(value: string | undefined);
    addEventReport(params: PartialBy<Omit<CallEventReport, 'serviceId' | 'roomId' | 'callId' | 'clientId' | 'userId' | 'peerConnectionId' | 'marker'>, 'timestamp'>): void;
    get closed(): boolean;
    close(): void;
    getTrack(trackId: string): ObservedInboundAudioTrack | ObservedInboundVideoTrack | ObservedOutboundAudioTrack | ObservedOutboundVideoTrack | undefined;
    resetMetrics(): void;
    update(sample: PeerConnectionTransport, timestamp: number): void;
    createInboundAudioTrack(config: ObservedInboundAudioTrackModel): ObservedInboundAudioTrack;
    createInboundVideoTrack(config: ObservedInboundVideoTrackModel): ObservedInboundVideoTrack;
    createOutboundAudioTrack(config: ObservedOutboundAudioTrackModel): ObservedOutboundAudioTrack;
    createOutboundVideoTrack(config: ObservedOutboundVideoTrackModel): ObservedOutboundVideoTrack;
    createDataChannel(channelId: number): ObservedDataChannel;
    updateMetrics(): void;
    private _updateQualityScore;
}
//# sourceMappingURL=ObservedPeerConnection.d.ts.map