"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObservedSfuTransport = void 0;
const events_1 = require("events");
const ObservedSfuInboundRtpPad_1 = require("./ObservedSfuInboundRtpPad");
const ObservedSfuOutboundRtpPad_1 = require("./ObservedSfuOutboundRtpPad");
const ObservedSfuSctpChannel_1 = require("./ObservedSfuSctpChannel");
class ObservedSfuTransport extends events_1.EventEmitter {
    constructor(_model, sfu) {
        super();
        this._model = _model;
        this.sfu = sfu;
        this.created = Date.now();
        this._updated = Date.now();
        this._closed = false;
        this._inboundRtpPads = new Map();
        this._outboundRtpPads = new Map();
        this._sctpChannels = new Map();
        this.setMaxListeners(Infinity);
    }
    get inboundRtpPads() {
        return this._inboundRtpPads;
    }
    get outboundRtpPads() {
        return this._outboundRtpPads;
    }
    get sctpChannels() {
        return this._sctpChannels;
    }
    get serviceId() {
        return this.sfu.serviceId;
    }
    get mediaUnitId() {
        return this.sfu.mediaUnitId;
    }
    get sfuId() {
        return this.sfu.sfuId;
    }
    get sfuTransportId() {
        return this._model.sfuTransportId;
    }
    get reports() {
        return this.sfu.reports;
    }
    get updated() {
        return this._updated;
    }
    get marker() {
        return this._marker;
    }
    set marker(value) {
        this._marker = value;
        this._inboundRtpPads.forEach((pad) => (pad.marker = value));
        this._outboundRtpPads.forEach((pad) => (pad.marker = value));
        this._sctpChannels.forEach((channel) => (channel.marker = value));
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this._closed)
            return;
        this._closed = true;
        [...this._inboundRtpPads.values()].forEach((pad) => pad.close());
        [...this._outboundRtpPads.values()].forEach((pad) => pad.close());
        [...this._sctpChannels.values()].forEach((channel) => channel.close());
        this.emit('close');
    }
    update(sample, timestamp) {
        const now = Date.now();
        const elapsedTimeInMs = now - this._updated;
        const report = Object.assign(Object.assign({ serviceId: this.serviceId, mediaUnitId: this.mediaUnitId, sfuId: this.sfuId, callId: undefined, roomId: undefined }, sample), { timestamp, marker: this.marker });
        this.reports.addSfuTransportReport(report);
        this.stats = sample;
        this._updated = now;
        this.emit('update', {
            elapsedTimeInMs,
        });
    }
    createSfuOutboundRtpPad(model) {
        if (this._closed)
            throw new Error('Cannot create SfuOutboundRtpPad on closed SfuTransport');
        if (this._inboundRtpPads.has(model.sfuOutboundRtpPadId))
            throw new Error(`SfuOutboundRtpPad with id ${model.sfuOutboundRtpPadId} already exists`);
        const result = new ObservedSfuOutboundRtpPad_1.ObservedSfuOutboundRtpPad(model, this);
        result.once('close', () => {
            this._outboundRtpPads.delete(model.sfuOutboundRtpPadId);
        });
        this._outboundRtpPads.set(model.sfuOutboundRtpPadId, result);
        this.emit('newsfuoutboundrtppad', result);
        return result;
    }
    createSfuInboundRtpPad(model) {
        if (this._closed)
            throw new Error('Cannot create SfuInboundRtpPad on closed SfuTransport');
        if (this._inboundRtpPads.has(model.sfuInboundRtpPadId))
            throw new Error(`SfuInboundRtpPad with id ${model.sfuInboundRtpPadId} already exists`);
        const result = new ObservedSfuInboundRtpPad_1.ObservedSfuInboundRtpPad(model, this);
        result.once('close', () => {
            this._inboundRtpPads.delete(model.sfuInboundRtpPadId);
        });
        this._inboundRtpPads.set(model.sfuInboundRtpPadId, result);
        this.emit('newsfuinboundrtppad', result);
        return result;
    }
    createSfuSctpChannel(model) {
        if (this._closed)
            throw new Error('Cannot create SfuSctpChannel on closed SfuTransport');
        if (this._sctpChannels.has(model.sfuSctpChannelId))
            throw new Error(`SfuSctpChannel with id ${model.sfuSctpChannelId} already exists`);
        const result = new ObservedSfuSctpChannel_1.ObservedSfuSctpChannel(model, this);
        result.once('close', () => {
            this._sctpChannels.delete(model.sfuSctpChannelId);
        });
        this._sctpChannels.set(model.sfuSctpChannelId, result);
        this.emit('newsfusctpchannel', result);
        return result;
    }
}
exports.ObservedSfuTransport = ObservedSfuTransport;
