"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Observer = void 0;
const logger_1 = require("./common/logger");
const ObservedCall_1 = require("./ObservedCall");
const ReportsCollector_1 = require("./ReportsCollector");
const events_1 = require("events");
const callEventReports_1 = require("./common/callEventReports");
const ObservedSfu_1 = require("./ObservedSfu");
const CallSummaryMonitor_1 = require("./monitors/CallSummaryMonitor");
const TurnUsageMonitor_1 = require("./monitors/TurnUsageMonitor");
const ClientIssueMonitor_1 = require("./monitors/ClientIssueMonitor");
const SfuServerMonitor_1 = require("./monitors/SfuServerMonitor");
const logger = (0, logger_1.createLogger)('Observer');
class Observer extends events_1.EventEmitter {
    static create(providedConfig) {
        const config = Object.assign({
            defaultServiceId: 'default-service-id',
            defaultMediaUnitId: 'default-media-unit-id',
            evaluator: {
                fetchSamples: true,
                maxIdleTimeInMs: 300 * 1000,
            },
            sink: {},
            logLevel: 'info',
        }, providedConfig);
        return new Observer(config);
    }
    constructor(config) {
        super();
        this.config = config;
        this.reports = new ReportsCollector_1.ReportsCollector();
        this._observedCalls = new Map();
        this._observedSfus = new Map();
        this._monitors = new Map();
        this._closed = false;
        this.setMaxListeners(Infinity);
        logger.debug('Observer is created with config', this.config);
        const onReports = (context) => this.emit('reports', context);
        const onNewReport = (collectedReports) => {
            if (!this.config.maxReports || this._closed)
                return;
            if (this.config.maxReports < collectedReports) {
                this._emitReports();
            }
        };
        this._emitReports();
        this.once('close', () => {
            this.reports.off('newreport', onNewReport);
            this.reports.off('reports', onReports);
        });
        this.reports.on('newreport', onNewReport);
        this.reports.on('reports', onReports);
    }
    createObservedCall(config) {
        if (this._closed) {
            throw new Error('Attempted to create a call source on a closed observer');
        }
        const { appData, reportCallEnded = true, reportCallStarted = true } = config, model = __rest(config, ["appData", "reportCallEnded", "reportCallStarted"]);
        const call = new ObservedCall_1.ObservedCall(Object.assign(Object.assign({}, model), { serviceId: this.config.defaultServiceId }), this, appData);
        if (this._closed)
            throw new Error('Cannot create an observed call on a closed observer');
        if (this._observedCalls.has(call.callId))
            throw new Error(`Observed Call with id ${call.callId} already exists`);
        call.once('close', () => {
            var _a;
            this._observedCalls.delete(call.callId);
            reportCallEnded && this.reports.addCallEventReport((0, callEventReports_1.createCallEndedEventReport)(call.serviceId, call.roomId, call.callId, (_a = call.observationEnded) !== null && _a !== void 0 ? _a : Date.now()));
        });
        this._observedCalls.set(call.callId, call);
        reportCallStarted && this.reports.addCallEventReport((0, callEventReports_1.createCallStartedEventReport)(call.serviceId, call.roomId, call.callId, call.observationStarted));
        this.emit('newcall', call);
        return call;
    }
    createObservedSfu(model, appData) {
        if (this._closed) {
            throw new Error('Attempted to create an sfu source on a closed observer');
        }
        const sfu = new ObservedSfu_1.ObservedSfu(model, this, appData);
        if (this._closed)
            throw new Error('Cannot create an observed sfu on a closed observer');
        if (this._observedSfus.has(sfu.sfuId))
            throw new Error(`Observed SFU with id ${sfu.sfuId} already exists`);
        sfu.once('close', () => {
            this._observedSfus.delete(sfu.sfuId);
        });
        this._observedSfus.set(sfu.sfuId, sfu);
        this.emit('newsfu', sfu);
        return sfu;
    }
    createCallSummaryMonitor(options) {
        if (this._closed)
            throw new Error('Cannot create a call summary monitor on a closed observer');
        const existingMonitor = this._monitors.get(CallSummaryMonitor_1.CallSummaryMonitor.name);
        if (existingMonitor)
            return existingMonitor;
        const monitor = new CallSummaryMonitor_1.CallSummaryMonitor(options);
        const onNewCall = (call) => {
            monitor.addCall(call);
            call.once('close', () => {
                var _a;
                return setTimeout(() => {
                    const summary = monitor.takeSummary(call.callId);
                    summary && monitor.emit('summary', summary);
                }, (_a = options === null || options === void 0 ? void 0 : options.timeoutAfterCallClose) !== null && _a !== void 0 ? _a : 1000);
            });
        };
        monitor.once('close', () => {
            this._monitors.delete(CallSummaryMonitor_1.CallSummaryMonitor.name);
            this.off('newcall', onNewCall);
        });
        this._monitors.set(CallSummaryMonitor_1.CallSummaryMonitor.name, monitor);
        this.on('newcall', onNewCall);
        this.once('close', () => {
            monitor.close();
        });
        return monitor;
    }
    createTurnUsageMonitor() {
        if (this._closed)
            throw new Error('Cannot create a turn usage monitor on a closed observer');
        const existingMonitor = this._monitors.get(TurnUsageMonitor_1.TurnUsageMonitor.name);
        if (existingMonitor)
            return existingMonitor;
        const monitor = new TurnUsageMonitor_1.TurnUsageMonitor();
        const onNewCall = (call) => {
            const onNewClient = (client) => {
                const onNewPeerConnection = (pc) => {
                    const onUsingTurnChanged = (usingTurn) => {
                        if (usingTurn)
                            monitor.addPeerConnection(pc);
                        else
                            monitor.removePeerConnection(pc);
                    };
                    pc.once('close', () => {
                        pc.ICE.off('usingturnchanged', onUsingTurnChanged);
                        monitor.removePeerConnection(pc);
                    });
                    pc.ICE.on('usingturnchanged', onUsingTurnChanged);
                };
                client.once('close', () => client.off('newpeerconnection', onNewPeerConnection));
                client.on('newpeerconnection', onNewPeerConnection);
            };
            call.once('close', () => call.off('newclient', onNewClient));
            call.on('newclient', onNewClient);
        };
        monitor.once('close', () => {
            this._monitors.delete(TurnUsageMonitor_1.TurnUsageMonitor.name);
            this.off('newcall', onNewCall);
        });
        this._monitors.set(TurnUsageMonitor_1.TurnUsageMonitor.name, monitor);
        this.on('newcall', onNewCall);
        this.once('close', () => {
            monitor.close();
        });
        return monitor;
    }
    createSfuServerMonitor() {
        if (this._closed)
            throw new Error('Cannot create a turn usage monitor on a closed observer');
        const existingMonitor = this._monitors.get(SfuServerMonitor_1.SfuServerMonitor.name);
        if (existingMonitor)
            return existingMonitor;
        const monitor = new SfuServerMonitor_1.SfuServerMonitor({
            tooHighRttAlertSettings: {
                minNumberOfPeerConnections: 10,
                percentageOfPeerConnectionsHighWatermark: 0.5,
                percentageOfPeerConnectionsLowWatermark: 0.2,
                threshold: 'rtt-gt-300',
            }
        });
        const onNewCall = (call) => {
            const onNewClient = (client) => {
                const onNewPeerConnection = (pc) => {
                    monitor.addPeerConnection(pc);
                };
                client.once('close', () => client.off('newpeerconnection', onNewPeerConnection));
                client.on('newpeerconnection', onNewPeerConnection);
            };
            call.once('close', () => call.off('newclient', onNewClient));
            call.on('newclient', onNewClient);
        };
        monitor.once('close', () => {
            this._monitors.delete(SfuServerMonitor_1.SfuServerMonitor.name);
            this.off('newcall', onNewCall);
        });
        this._monitors.set(SfuServerMonitor_1.SfuServerMonitor.name, monitor);
        this.on('newcall', onNewCall);
        this.once('close', () => {
            monitor.close();
        });
        return monitor;
    }
    createClientIssueMonitor() {
        if (this._closed)
            throw new Error('Cannot create a turn usage monitor on a closed observer');
        const existingMonitor = this._monitors.get(ClientIssueMonitor_1.ClientIssueMonitor.name);
        if (existingMonitor)
            return existingMonitor;
        const monitor = new ClientIssueMonitor_1.ClientIssueMonitor();
        const onNewCall = (call) => monitor.addCall(call);
        monitor.once('close', () => {
            this._monitors.delete(ClientIssueMonitor_1.ClientIssueMonitor.name);
            this.off('newcall', onNewCall);
        });
        this._monitors.set(ClientIssueMonitor_1.ClientIssueMonitor.name, monitor);
        this.on('newcall', onNewCall);
        this.once('close', () => {
            monitor.close();
        });
        return monitor;
    }
    get observedCalls() {
        return this._observedCalls;
    }
    get observedSfus() {
        return this._observedSfus;
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this._closed) {
            return logger.debug('Attempted to close twice');
        }
        this._closed = true;
        this._observedCalls.forEach((call) => call.close());
        this.emit('close');
    }
    _emitReports() {
        if (this._reportTimer) {
            clearTimeout(this._reportTimer);
            this._reportTimer = undefined;
        }
        this.reports.emit();
        if (this.config.maxCollectingTimeInMs) {
            this._reportTimer = setTimeout(() => {
                this._reportTimer = undefined;
                this._emitReports();
            }, this.config.maxCollectingTimeInMs);
        }
    }
}
exports.Observer = Observer;
