"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CallSummaryMonitor = void 0;
const events_1 = require("events");
class CallSummaryMonitor extends events_1.EventEmitter {
    constructor(config = {}) {
        super();
        this.config = config;
        this._summaries = new Map();
        this._closed = false;
        this.setMaxListeners(Infinity);
    }
    addCall(call) {
        if (this.closed)
            return;
        let callSummary = this._summaries.get(call.callId);
        if (!callSummary) {
            callSummary = {
                callId: call.callId,
                roomId: call.roomId,
                serviceId: call.serviceId,
                clients: [],
                durationInMs: 0,
                maxNumberOfParticipants: 0,
                numberOfIssues: 0,
                started: call.created,
            };
            this._summaries.set(call.callId, callSummary);
        }
        const constCallSummary = callSummary;
        const onNewClient = (client) => {
            constCallSummary.maxNumberOfParticipants = Math.max(constCallSummary.maxNumberOfParticipants, call.clients.size);
            this._addClient(constCallSummary, client);
        };
        call.once('close', () => {
            call.off('newclient', onNewClient);
        });
        call.on('newclient', onNewClient);
    }
    _addClient(callSummary, client) {
        if (this.closed)
            return;
        const clientSummary = {
            clientId: client.clientId,
            mediaUnitId: client.mediaUnitId,
            durationInMs: 0,
            totalInboundAudioBytes: 0,
            totalInboundVideoBytes: 0,
            totalOutboundAudioBytes: 0,
            totalOutboundVideoBytes: 0,
            totalDataChannelBytesSent: 0,
            totalDataChannelBytesReceived: 0,
            ewmaRttInMs: 0,
            joined: client.created,
            userId: client.userId,
            usedTurn: false,
            issues: [],
        };
        callSummary.clients.push(clientSummary);
        const updateClient = () => {
            if (client.avgRttInMs) {
                clientSummary.ewmaRttInMs = (clientSummary.ewmaRttInMs * 0.9) + (client.avgRttInMs * 0.1);
            }
        };
        const onUsingTurn = (usingTurn) => {
            clientSummary.usedTurn || (clientSummary.usedTurn = usingTurn);
        };
        const onUserMediaError = (error) => {
            if (!this.config.detectUserMediaIssues)
                return;
            client.addIssue({
                severity: 'critical',
                timestamp: Date.now(),
                description: error,
            });
        };
        const onNewPeerConnection = (peerConnection) => this._addPeerConnection(clientSummary, peerConnection);
        client.on('update', updateClient);
        client.on('usingturn', onUsingTurn);
        client.on('usermediaerror', onUserMediaError);
        client.on('newpeerconnection', onNewPeerConnection);
        client.once('close', () => {
            const now = Date.now();
            clientSummary.totalInboundAudioBytes = client.totalReceivedAudioBytes;
            clientSummary.totalInboundVideoBytes = client.totalReceivedVideoBytes;
            clientSummary.totalOutboundAudioBytes = client.totalSentAudioBytes;
            clientSummary.totalOutboundVideoBytes = client.totalSentVideoBytes;
            clientSummary.totalDataChannelBytesSent = client.totalDataChannelBytesSent;
            clientSummary.totalDataChannelBytesReceived = client.totalDataChannelBytesReceived;
            clientSummary.durationInMs = now - client.created;
            clientSummary.left = now;
            // client does not store issues after update emitted
            // clientSummary.issues.push(...client.issues);
            client.off('update', updateClient);
            client.off('usingturn', onUsingTurn);
            client.off('usermediaerror', onUserMediaError);
            client.off('newpeerconnection', onNewPeerConnection);
            callSummary.durationInMs += clientSummary.durationInMs;
            callSummary.numberOfIssues += clientSummary.issues.length;
            callSummary.highestSeverity = clientSummary.issues.reduce((highest, issue) => {
                if (issue.severity === 'critical')
                    return 'critical';
                if (issue.severity === 'major' && highest !== 'critical')
                    return 'major';
                if (issue.severity === 'minor' && highest !== 'critical' && highest !== 'major')
                    return 'minor';
                return highest;
            }, callSummary.highestSeverity);
        });
    }
    _addPeerConnection(clientSummary, peerConnection) {
        const onOutboundVideoTrack = (track) => this._addOutboundVideoTrack(clientSummary, track);
        peerConnection.on('newoutboundvideotrack', onOutboundVideoTrack);
        peerConnection.once('close', () => {
            peerConnection.off('newoutboundvideotrack', onOutboundVideoTrack);
        });
    }
    _addOutboundVideoTrack(clientSummary, track) {
        const onQualityLimitationChanged = (reason) => {
            if (!this.config.detectMediaTrackQualityLimitationIssues)
                return;
            track.peerConnection.client.addIssue({
                severity: 'minor',
                timestamp: Date.now(),
                description: reason,
                peerConnectionId: track.peerConnectionId,
                trackId: track.trackId,
            });
        };
        track.once('close', () => {
            track.off('qualitylimitationchanged', onQualityLimitationChanged);
        });
        track.on('qualitylimitationchanged', onQualityLimitationChanged);
    }
    takeSummary(callId) {
        if (this.closed)
            return;
        const summary = this._summaries.get(callId);
        this._summaries.delete(callId);
        return summary;
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this._closed)
            return;
        this._closed = true;
        this.emit('close');
    }
}
exports.CallSummaryMonitor = CallSummaryMonitor;
