"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SfuServerMonitor = void 0;
const events_1 = require("events");
class SfuServerMonitor extends events_1.EventEmitter {
    constructor(config) {
        super();
        this.config = config;
        this._closed = false;
        this._peerConnections = new Map();
        this._metrics = new Map();
        this._rttAlertsOn = new Set();
    }
    get stats() {
        return Array.from(this._metrics.values());
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this.closed)
            return;
        this._closed = true;
        this._peerConnections.clear();
        this._metrics.clear();
        this._rttAlertsOn.clear();
        this.emit('close');
    }
    addPeerConnection(peerConnection) {
        if (this.closed)
            return;
        let registered = false;
        const onUpdate = () => {
            const sfuId = peerConnection.client.sfuId;
            const peerConnectionRecord = this._peerConnections.get(peerConnection.peerConnectionId);
            if (!sfuId)
                return;
            let metrics = this._metrics.get(sfuId);
            if (!metrics) {
                metrics = {
                    sfuId,
                    receivedPacketsPerSecond: 0,
                    sentPacketsPerSecond: 0,
                    receivingAudioBitrate: 0,
                    receivingVideoBitrate: 0,
                    sendingAudioBitrate: 0,
                    sendingVideoBitrate: 0,
                    numberOfPeerConnectionsRttLt150: 0,
                    numberOfPeerConnectionsRttLt300: 0,
                    numberOfPeerConnectionsRttLt50: 0,
                    numberOfPeerConnectionsRttOver300: 0,
                    numberOfPeerConnections: 0,
                };
                this._metrics.set(sfuId, metrics);
            }
            if (!registered) {
                registered = true;
                ++metrics.numberOfPeerConnections;
            }
            if (peerConnectionRecord) {
                this._subtract(metrics, peerConnectionRecord);
            }
            this._increase(metrics, peerConnection);
            this._peerConnections.set(peerConnection.peerConnectionId, {
                lastSendingAudioBitrate: peerConnection.sendingAudioBitrate,
                lastSendingVideoBitrate: peerConnection.sendingVideoBitrate,
                lastReceivingAudioBitrate: peerConnection.receivingAudioBitrate,
                lastReceivingVideoBitrate: peerConnection.receivingVideoBitrate,
                lastAvgRttInMs: peerConnection.avgRttInMs,
                lastReceivingPacketsPerSecond: peerConnection.receivingPacketsPerSecond,
                lastSendingPacketsPerSecond: peerConnection.sendingPacketsPerSecond,
            });
            this._checkHighRttForClients(metrics);
        };
        peerConnection.once('close', () => {
            var _a, _b;
            const address = (_b = (_a = peerConnection.ICE.selectedRemoteCandidate) === null || _a === void 0 ? void 0 : _a.address) !== null && _b !== void 0 ? _b : '';
            const peerConnectionRecord = this._peerConnections.get(peerConnection.peerConnectionId);
            const metrics = this._metrics.get(address);
            if (metrics) {
                if (peerConnectionRecord) {
                    this._subtract(metrics, peerConnectionRecord);
                }
                if (registered) {
                    if (--metrics.numberOfPeerConnections === 0) {
                        this._metrics.delete(address);
                        this._rttAlertsOn.delete(address);
                    }
                }
            }
            this._peerConnections.delete(peerConnection.peerConnectionId);
            peerConnection.off('update', onUpdate);
        });
        peerConnection.on('update', onUpdate);
    }
    _checkHighRttForClients(metrics) {
        const totalClients = metrics.numberOfPeerConnections;
        const { percentageOfPeerConnectionsHighWatermark, percentageOfPeerConnectionsLowWatermark, minNumberOfPeerConnections, threshold } = this.config.tooHighRttAlertSettings;
        if (metrics.numberOfPeerConnections < minNumberOfPeerConnections)
            return;
        const rttLt150Percentage = (metrics.numberOfPeerConnectionsRttLt150 + metrics.numberOfPeerConnectionsRttLt300 + metrics.numberOfPeerConnectionsRttOver300) / totalClients;
        const rttLt300Percentage = (metrics.numberOfPeerConnectionsRttLt150 + metrics.numberOfPeerConnectionsRttLt300) / totalClients;
        const rttOver300Percentage = metrics.numberOfPeerConnectionsRttOver300 / totalClients;
        const alertOn = this._rttAlertsOn.has(metrics.sfuId);
        switch (threshold) {
            case 'rtt-gt-50': {
                if (!alertOn && percentageOfPeerConnectionsHighWatermark < rttLt150Percentage) {
                    this.emit('high-rtt-started', {
                        sfuId: metrics.sfuId,
                        threshold: 'rtt-gt-50',
                        percentageOfPeerConnections: rttLt150Percentage
                    });
                    this._rttAlertsOn.add(metrics.sfuId);
                }
                else if (alertOn && rttLt150Percentage < percentageOfPeerConnectionsLowWatermark) {
                    this.emit('high-rtt-cleared', { sfuId: metrics.sfuId });
                    this._rttAlertsOn.delete(metrics.sfuId);
                }
                break;
            }
            case 'rtt-gt-150': {
                if (!alertOn && percentageOfPeerConnectionsHighWatermark < rttLt300Percentage) {
                    this.emit('high-rtt-started', {
                        sfuId: metrics.sfuId,
                        threshold: 'rtt-gt-150',
                        percentageOfPeerConnections: rttLt300Percentage
                    });
                    this._rttAlertsOn.add(metrics.sfuId);
                }
                else if (alertOn && rttLt300Percentage < percentageOfPeerConnectionsLowWatermark) {
                    this.emit('high-rtt-cleared', { sfuId: metrics.sfuId });
                    this._rttAlertsOn.delete(metrics.sfuId);
                }
                break;
            }
            case 'rtt-gt-300': {
                if (!alertOn && percentageOfPeerConnectionsHighWatermark < rttOver300Percentage) {
                    this.emit('high-rtt-started', {
                        sfuId: metrics.sfuId,
                        threshold: 'rtt-gt-300',
                        percentageOfPeerConnections: rttOver300Percentage
                    });
                    this._rttAlertsOn.add(metrics.sfuId);
                }
                else if (alertOn && rttOver300Percentage < percentageOfPeerConnectionsLowWatermark) {
                    this.emit('high-rtt-cleared', { sfuId: metrics.sfuId });
                    this._rttAlertsOn.delete(metrics.sfuId);
                }
                break;
            }
        }
    }
    _increase(metrics, peerConnection) {
        metrics.sendingAudioBitrate += peerConnection.receivingAudioBitrate;
        metrics.sendingVideoBitrate += peerConnection.receivingVideoBitrate;
        metrics.receivingAudioBitrate += peerConnection.sendingAudioBitrate;
        metrics.receivingVideoBitrate += peerConnection.sendingVideoBitrate;
        metrics.receivedPacketsPerSecond += peerConnection.sendingPacketsPerSecond;
        metrics.sentPacketsPerSecond += peerConnection.receivingPacketsPerSecond;
        if (peerConnection.avgRttInMs) {
            if (peerConnection.avgRttInMs < 50) {
                ++metrics.numberOfPeerConnectionsRttLt50;
            }
            else if (peerConnection.avgRttInMs < 150) {
                ++metrics.numberOfPeerConnectionsRttLt150;
            }
            else if (peerConnection.avgRttInMs < 300) {
                ++metrics.numberOfPeerConnectionsRttLt300;
            }
            else {
                ++metrics.numberOfPeerConnectionsRttOver300;
            }
        }
    }
    _subtract(metrics, peerConnectionRecord) {
        metrics.sendingAudioBitrate -= peerConnectionRecord.lastReceivingAudioBitrate;
        metrics.sendingVideoBitrate -= peerConnectionRecord.lastReceivingVideoBitrate;
        metrics.receivingAudioBitrate -= peerConnectionRecord.lastSendingAudioBitrate;
        metrics.receivingVideoBitrate -= peerConnectionRecord.lastSendingVideoBitrate;
        metrics.receivedPacketsPerSecond -= peerConnectionRecord.lastSendingPacketsPerSecond;
        metrics.sentPacketsPerSecond -= peerConnectionRecord.lastReceivingPacketsPerSecond;
        if (peerConnectionRecord.lastAvgRttInMs) {
            if (peerConnectionRecord.lastAvgRttInMs < 50) {
                --metrics.numberOfPeerConnectionsRttLt50;
            }
            else if (peerConnectionRecord.lastAvgRttInMs < 150) {
                --metrics.numberOfPeerConnectionsRttLt150;
            }
            else if (peerConnectionRecord.lastAvgRttInMs < 300) {
                --metrics.numberOfPeerConnectionsRttLt300;
            }
            else {
                --metrics.numberOfPeerConnectionsRttOver300;
            }
        }
    }
}
exports.SfuServerMonitor = SfuServerMonitor;
