"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TurnUsageMonitor = void 0;
const events_1 = require("events");
class TurnUsageMonitor extends events_1.EventEmitter {
    constructor() {
        super();
        this._connections = new Map();
        this._turnUsage = new Map();
        this._stats = new Map();
        this._closed = false;
        this.setMaxListeners(Infinity);
    }
    get stats() {
        var _a;
        const result = [];
        for (const [turnIp, usage] of this._turnUsage) {
            const stats = (_a = this._stats.get(turnIp)) !== null && _a !== void 0 ? _a : {
                turnIp,
                totalBytesSent: 0,
                totalBytesReceived: 0,
                totalPacketsSent: 0,
                totalPacketsReceived: 0,
                deltaBytesSent: 0,
                deltaBytesReceived: 0,
                timestamp: Date.now(),
                outboundBitrate: 0,
                inboundBitrate: 0,
            };
            stats.deltaBytesSent = usage.totalBytesSent - stats.totalBytesSent;
            stats.deltaBytesReceived = usage.totalBytesReceived - stats.totalBytesReceived;
            stats.timestamp = Date.now();
            stats.outboundBitrate = (stats.deltaBytesSent * 8) / ((stats.timestamp - stats.timestamp) / 1000.0);
            stats.inboundBitrate = (stats.deltaBytesReceived * 8) / ((stats.timestamp - stats.timestamp) / 1000.0);
            stats.totalBytesSent = usage.totalBytesSent;
            stats.totalBytesReceived = usage.totalBytesReceived;
            stats.totalPacketsSent = usage.totalPacketsSent;
            stats.totalPacketsReceived = usage.totalPacketsReceived;
            this._stats.set(turnIp, stats);
            result.push(stats);
        }
        return result;
    }
    addPeerConnection(peerConnection) {
        if (this.closed)
            return;
        if (this._connections.has(peerConnection.peerConnectionId))
            return;
        const ice = peerConnection.ICE;
        const onUpdate = () => {
            var _a;
            const turnIp = (_a = ice === null || ice === void 0 ? void 0 : ice.selectedRemoteCandidate) === null || _a === void 0 ? void 0 : _a.address;
            if (!peerConnection.usingTURN || !turnIp)
                return;
            let usage = this._turnUsage.get(turnIp);
            if (!usage) {
                usage = {
                    turnIp,
                    totalBytesSent: 0,
                    totalBytesReceived: 0,
                    totalPacketsSent: 0,
                    totalPacketsReceived: 0,
                };
                this._turnUsage.set(turnIp, usage);
                this.emit('newturnip', turnIp);
            }
            usage.totalBytesSent += ice.deltaBytesSent;
            usage.totalBytesReceived += ice.deltaBytesReceived;
            usage.totalPacketsSent += ice.deltaPacketsSent;
            usage.totalPacketsReceived += ice.deltaPacketsReceived;
            this.emit('update', usage);
        };
        peerConnection.ICE.on('update', onUpdate);
        this._connections.set(peerConnection.peerConnectionId, {
            peerConnection,
            onUpdate
        });
        this.emit('addpeerconnection', peerConnection);
    }
    removePeerConnection(peerConnection) {
        var _a;
        if (this.closed)
            return;
        if (!this._connections.has(peerConnection.peerConnectionId))
            return;
        const { onUpdate } = (_a = this._connections.get(peerConnection.peerConnectionId)) !== null && _a !== void 0 ? _a : {};
        if (!onUpdate)
            return;
        peerConnection.ICE.off('update', onUpdate);
        this._connections.delete(peerConnection.peerConnectionId);
        this.emit('removepeerconnection', peerConnection);
    }
    get turnIps() {
        return Array.from(this._turnUsage.keys());
    }
    get peerConnections() {
        return Array.from(this._connections.values()).map(({ peerConnection }) => peerConnection);
    }
    getUsage(turnIp) {
        return this._turnUsage.get(turnIp);
    }
    get clients() {
        const result = new Map();
        for (const pc of this.peerConnections) {
            if (result.has(pc.clientId))
                continue;
            result.set(pc.clientId, pc.client);
        }
        return result;
    }
    get closed() {
        return this._closed;
    }
    close() {
        if (this._closed)
            return;
        this._closed = true;
        this.emit('close');
    }
}
exports.TurnUsageMonitor = TurnUsageMonitor;
